"""teilen flask-backend configuration."""

import os
from pathlib import Path
from uuid import uuid4


# pylint: disable=R0903:too-few-public-methods
class AppConfig:
    """teilen-backend configuration."""

    MODE = os.environ.get("MODE", "prod")  # "prod" | "dev"
    DEV_CORS_FRONTEND_URL = os.environ.get(
        "DEV_CORS_FRONTEND_URL", "http://localhost:3000"
    )
    PORT = os.environ.get("PORT", "27183" if MODE == "prod" else "5000")
    FLASK_THREADS = 5
    GUNICORN_OPTIONS = None

    STATIC_PATH = Path(__file__).parent / "client"
    SESSION_COOKIE_NAME = "teilen_session"
    SECRET_KEY = os.environ.get("SECRET_KEY", str(uuid4()))
    WORKING_DIR = Path(os.environ.get("WORKING_DIR", Path.cwd())).resolve()
    PASSWORD = os.environ.get("PASSWORD")
    ARCHIVE_BUILD_CONCURRENCY = int(
        os.environ.get("ARCHIVE_BUILD_CONCURRENCY", 3)
    )



import os
from pathlib import Path
from setuptools import setup


try:
    long_description = (Path(__file__).parent.parent / "README.md").read_text(
        encoding="utf8"
    )
except FileNotFoundError:
    long_description = "See docs at https://github.com/RichtersFinger/teilen"


setup(
    version=os.environ.get("VERSION", "0.6.1"),
    name="teilen",
    description="a simple application to share data via http with python flask backend and react frontend",
    long_description=long_description,
    long_description_content_type="text/markdown",
    author="Steffen Richters-Finger",
    author_email="srichters@uni-muenster.de",
    license="MIT",
    url="https://pypi.org/project/teilen/",
    project_urls={"Source": "https://github.com/RichtersFinger/python-teilen"},
    python_requires=">=3.10",
    install_requires=[
        "Flask>=3,<4",
        "gunicorn",
    ],
    packages=[
        "teilen",
    ],
    package_data={"teilen": ["client/**/*"]},
    entry_points={
        "console_scripts": [
            "teilen = teilen.app:run",
        ],
    },
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Intended Audience :: End Users/Desktop",
        "Topic :: Communications",
        "Topic :: Communications :: File Sharing",
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Programming Language :: Python :: 3.12",
        "Framework :: Flask",
    ],
)



"""teilen-backend definition."""

import sys
from pathlib import Path
import socket
import urllib.request
from importlib.metadata import version

from flask import (
    Flask,
    Response,
    send_from_directory,
)

from teilen.config import AppConfig
from teilen.api import register_api


def load_cors(_app: Flask, url: str) -> None:
    """Loads CORS-extension if required."""
    try:
        # pylint: disable=import-outside-toplevel
        from flask_cors import CORS
    except ImportError:
        print(
            "\033[31mERROR: Missing 'Flask-CORS'-package for dev-server. "
            + "Install with 'pip install flask-cors'.\033[0m",
            file=sys.stderr,
        )
        sys.exit(1)
    else:
        print("INFO: Configuring app for CORS.", file=sys.stderr)
        _ = CORS(
            _app,
            supports_credentials=True,
            resources={"*": {"origins": url}},
        )


def load_callback_url_options() -> list[dict]:
    """
    Returns a list of IP-addresses with a name.

    Every record contains the fields 'name' and 'address'.
    """
    options = []

    # get LAN-address (https://stackoverflow.com/a/28950776)
    s = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
    s.settimeout(0)
    try:
        s.connect(("10.254.254.254", 1))
        options.append(
            {"address": "http://" + s.getsockname()[0], "name": "local"}
        )
    # pylint: disable=broad-exception-caught
    except Exception:
        pass
    finally:
        s.close()

    # get global IP
    try:
        with urllib.request.urlopen(
            "https://api.ipify.org", timeout=1
        ) as response:
            options.append(
                {
                    "address": "http://" + response.read().decode("utf-8"),
                    "name": "global",
                }
            )
    # pylint: disable=broad-exception-caught
    except Exception:
        pass

    return options


def print_welcome_message(config: AppConfig) -> None:
    """Prints welcome message to stdout."""
    url_options = load_callback_url_options()
    lines = (
        (["Running in dev-mode."] if config.MODE == "dev" else [])
        + [
            "Your teilen-instance will be available shortly.",
            "",
            "The contents of the following directory will be available:",
            (
                str(config.WORKING_DIR)[:30]
                + "..."
                + str(config.WORKING_DIR)[-30:]
                if len(str(config.WORKING_DIR)) > 70
                else str(config.WORKING_DIR)
            ),
        ]
        + (
            ["Password protection is active."]
            if config.PASSWORD is not None
            else []
        )
        + (
            ["", "The following addresses have been detected automatically:"]
            if url_options
            else []
        )
        + list(
            map(
                lambda o: f" * {o['name']}: {o['address']}:{config.PORT}",
                url_options,
            )
        )
    )
    delimiter = "#" * (max(map(len, lines)) + 4)
    print(delimiter)
    for line in lines:
        print(f"# {line}{' '*(len(delimiter) - len(line) - 4)} #")
    print(delimiter)


def app_factory(config: AppConfig) -> Flask:
    """Returns teilen-Flask app."""
    if not config.WORKING_DIR.is_dir():
        print(
            "\033[1;31mERROR\033[0m: "
            + f"Requested directory '{config.WORKING_DIR}' does not exist.",
            file=sys.stderr,
        )
        sys.exit(1)
    if not config.WORKING_DIR.is_absolute():
        config.WORKING_DIR = config.WORKING_DIR.resolve()

    # define Flask-app
    _app = Flask(__name__, static_folder=config.STATIC_PATH)

    _app.config.from_object(config)

    # extensions
    if config.MODE == "dev":
        load_cors(_app, config.DEV_CORS_FRONTEND_URL)

    # print welcome message
    print_welcome_message(config)

    @_app.route("/ping", methods=["GET"])
    def ping():
        """
        Returns 'pong'.
        """
        return Response("pong", mimetype="text/plain", status=200)

    @_app.route("/version", methods=["GET"])
    def get_version():
        """
        Returns app version.
        """
        return Response(version("teilen"), mimetype="text/plain", status=200)

    register_api(_app, config)

    @_app.route("/", defaults={"path": ""})
    @_app.route("/<path:path>")
    def get_client(path):
        """Serve static content."""
        if path != "":
            return send_from_directory(config.STATIC_PATH, path)
        return send_from_directory(config.STATIC_PATH, "index.html")

    return _app


def parse_cmdline_args(config: AppConfig):
    """Update config using command line arguments."""

    if "-h" in sys.argv or "--help" in sys.argv:
        print(f"""Open a teilen-share
Software version: {version("teilen")}

Usage: teilen [options] [path]

Options:
  -h, --help                        Output this message and exit.
  -p, --password                    Set a password-requirement for this
                                    share.
  --port                            Set a specific port to run on.
                                    [Default {config.PORT}]

Arguments:
  path                              path to the directory that is shared
                                    [Default current working directory]
""", end="")
        sys.exit(0)

    index = 1
    while True:
        if index >= len(sys.argv):
            break

        # password
        if sys.argv[index] in ["-p", "--password"]:
            if len(sys.argv) <= index + 1:
                print(
                    "\033[1;31mERROR\033[0m: "
                    + f"Missing value for option '{sys.argv[index]}'.",
                    file=sys.stderr,
                )
                sys.exit(1)
            config.PASSWORD = sys.argv[index + 1]
            index += 2
            continue

        # port
        if sys.argv[index] in ["--port"]:
            if len(sys.argv) <= index + 1:
                print(
                    "\033[1;31mERROR\033[0m: "
                    + f"Missing value for option '{sys.argv[index]}'.",
                    file=sys.stderr,
                )
                sys.exit(1)
            config.PORT = sys.argv[index + 1]
            index += 2
            continue

        # working directory (should by the last and a singular argument)
        if len(sys.argv) > index + 1:
            print(
                "\033[1;31mERROR\033[0m: "
                + f"Unknown option '{sys.argv[index]}'.",
                file=sys.stderr,
            )
            sys.exit(1)
        config.WORKING_DIR = Path(sys.argv[index]).resolve()
        index += 1


def run(app=None, config=None):
    """Run flask-app."""
    # load default config
    if not config:
        config = AppConfig()

    # parse command line arguments
    parse_cmdline_args(config)

    # load default app
    if not app:
        app = app_factory(config)

    # not intended for production due to, e.g., cors
    if config.MODE != "prod":
        print(
            "\033[1;33mWARNING\033[0m: "
            + f"Running in unexpected MODE '{config.Mode}'.",
            file=sys.stderr,
        )

    # prioritize gunicorn over werkzeug
    try:
        import gunicorn.app.base
    except ImportError:
        print(
            "\033[1;33mWARNING\033[0m: "
            + "Running without proper wsgi-server.",
            file=sys.stderr,
        )
        app.run(host="0.0.0.0", port=config.PORT)
    else:

        class StandaloneApplication(gunicorn.app.base.BaseApplication):
            """See https://docs.gunicorn.org/en/stable/custom.html"""

            def __init__(self, app_, options=None):
                self.options = options or {}
                self.application = app_
                super().__init__()

            def load_config(self):
                _config = {
                    key: value
                    for key, value in self.options.items()
                    if key in self.cfg.settings and value is not None
                }
                for key, value in _config.items():
                    self.cfg.set(key.lower(), value)

            def load(self):
                return self.application

        def post_worker_init(worker):
            """
            This removes atexit-handlers of multiprocessing that should
            not be run in worker-processes.

            See https://github.com/benoitc/gunicorn/issues/1391
            """
            try:
                # pylint: disable=import-outside-toplevel
                import atexit
                from multiprocessing.util import _exit_function
                atexit.unregister(_exit_function)
            # pylint: disable=broad-exception-caught
            except Exception:
                pass

        StandaloneApplication(
            app,
            {
                "bind": f"0.0.0.0:{config.PORT}",
                "workers": 1,
                "threads": config.FLASK_THREADS,
                "post_worker_init": post_worker_init,
            }
            | (config.GUNICORN_OPTIONS or {}),
        ).run()



"""teilen api-definition"""

from typing import Optional
from functools import wraps
from pathlib import Path
from urllib.parse import unquote
from datetime import datetime
from tempfile import mkdtemp
import zipfile
from uuid import uuid4
import multiprocessing
from shutil import rmtree
import atexit

from flask import (
    Flask,
    Response,
    jsonify,
    request,
    session,
    send_file,
    send_from_directory,
)

from teilen.config import AppConfig


def login_required(password: Optional[str]):
    """
    Protect endpoint with auth via 'teilen_session'-cookie or
    'X-Teilen-Auth'-header.
    """

    def decorator(route):
        @wraps(route)
        def __():
            if (
                session.get("password", request.headers.get("X-Teilen-Auth"))
                != password
            ):
                return Response("FAILED", mimetype="text/plain", status=401)
            return route()

        return __

    return decorator


def create_archive(
    id_: str,
    source: Path,
    tmp_dir: Path,
    progress,  # multiprocessing.managers.SyncManager.dict
):
    """
    Builds an archive from the data in `source` in
    `tmp_dir/id_/(source.name + ".zip")`. Progress updates are written
    to synced dict `progress`.
    """
    progress["status"] = "running"
    working_dir: Path = tmp_dir / id_
    working_dir.mkdir()
    destination: Path = working_dir / (source.name + ".zip")
    print(
        f"[{datetime.now().isoformat()} - {id_}] Creating archive for "
        + f"'{source}' in '{destination}'."
    )

    # create archive
    files = list(filter(Path.is_file, source.glob("**/*")))
    progress["totalFiles"] = len(files)
    progress["processedFiles"] = 0
    with zipfile.ZipFile(destination, "w", zipfile.ZIP_STORED) as archive:
        for f in files:
            progress["currentFile"] = str(
                f.resolve().relative_to(source.parent.resolve())
            )
            archive.write(
                f,
                f.resolve().relative_to(source.parent.resolve()),
            )
            progress["archiveSize"] = destination.stat().st_size
            progress["processedFiles"] += 1

    progress["currentFile"] = None
    # finalize
    print(
        f"[{datetime.now().isoformat()} - {id_}] Building archive successful."
    )
    progress["artifact"] = destination.name
    progress["status"] = "completed"


def silentfail(callback):
    """
    Returns a callable that run the given callback while silently
    capturing all exceptions.
    """

    def _(*args, **kwargs):
        try:
            callback(*args, **kwargs)
        # pylint: disable=broad-exception-caught
        except Exception:
            pass

    return _


def register_api(app: Flask, config: AppConfig):
    """Sets up api endpoints."""

    archive_store = {}
    mp_manager = multiprocessing.Manager()  # inter-process communication
    # use silentfail to avoid errors due to gunicorn master + workers
    # running this function
    atexit.register(silentfail(mp_manager.shutdown))

    tmp_dir = Path(mkdtemp(prefix="teilen-")).resolve()
    # use silentfail to avoid errors due to gunicorn master + workers
    # running this function
    atexit.register(silentfail(rmtree), tmp_dir)

    @app.route("/configuration", methods=["GET"])
    def get_configuration():
        """
        Get basic info on configuration.
        """
        return jsonify({"passwordRequired": config.PASSWORD is not None}), 200

    @app.route("/login", methods=["GET"])
    @login_required(config.PASSWORD)
    def get_login():
        """Test login."""
        session["password"] = config.PASSWORD
        return Response("OK", mimetype="text/plain", status=200)

    def get_location(provide_default: bool = True) -> Optional[Path]:
        """Parse and return location-arg."""
        if request.args.get("location") is None:
            if provide_default:
                return config.WORKING_DIR
            return None
        return (
            config.WORKING_DIR / unquote(request.args["location"])
        ).resolve()

    @app.route("/contents", methods=["GET"])
    @login_required(config.PASSWORD)
    def get_contents():
        """
        Returns contents for given location.
        """
        _location = get_location()

        # check for problems
        if (
            config.WORKING_DIR != _location
            and config.WORKING_DIR not in _location.parents
        ):
            return Response("Not allowed.", mimetype="text/plain", status=403)
        if not _location.is_dir():
            return Response(
                "Does not exist.", mimetype="text/plain", status=404
            )

        contents = list(_location.glob("*"))
        folders = filter(lambda p: p.is_dir(), contents)
        files = filter(lambda p: p.is_file(), contents)
        return (
            jsonify(
                [
                    {
                        "type": "folder",
                        "name": f.name,
                        "mtime": f.stat().st_mtime,
                    }
                    for f in folders
                ]
                + [
                    {
                        "type": "file",
                        "name": f.name,
                        "mtime": f.stat().st_mtime,
                        "size": f.stat().st_size,
                    }
                    for f in files
                ]
            ),
            200,
        )

    @app.route("/content", methods=["GET"])
    @login_required(config.PASSWORD)
    def get_content():
        """
        Returns file for given location or an id for an archive-job.
        """
        _location = get_location(False)
        # if forStatus, client only wants to validate request
        for_status = request.args.get("forStatus") is not None

        if _location is None:
            return Response(
                "Missing 'location' arg.", mimetype="text/plain", status=400
            )

        # check for problems
        if config.WORKING_DIR not in _location.parents:
            return Response("Not allowed.", mimetype="text/plain", status=403)
        if not _location.exists():
            return Response(
                "Does not exist.", mimetype="text/plain", status=404
            )

        if _location.is_file():
            if for_status:
                return Response("OK", mimetype="text/plain", status=200)
            return send_from_directory(
                config.WORKING_DIR,
                _location.relative_to(config.WORKING_DIR),
                as_attachment=True,
            )

        # create asynchronous job to build archive
        if _location.is_dir():
            # check already running jobs
            n_jobs = 0
            for job in archive_store.values():
                if "process" in job and job["process"].is_alive():
                    n_jobs += 1
            if n_jobs > config.ARCHIVE_BUILD_CONCURRENCY:
                return Response("BUSY", mimetype="text/plain", status=503)

            # run job
            # no threading-lock needed here because of uuids
            job_id = str(uuid4())
            progress = mp_manager.dict(
                {
                    "status": "queued",
                }
            )
            archive_store[job_id] = {
                "id": job_id,
                "progress": progress,
                "process": multiprocessing.Process(
                    target=create_archive,
                    args=(
                        job_id,
                        _location.resolve(),
                        tmp_dir,
                        progress,
                    ),
                    daemon=True,
                ),
            }
            archive_store[job_id]["process"].start()

            return jsonify({"id": job_id}), 202

        return Response("Unkown type.", mimetype="text/plain", status=501)

    @app.route("/archive-progress", methods=["GET"])
    @login_required(config.PASSWORD)
    def get_archive_progress():
        """Returns progress of an archive-job if available."""
        job_id = request.args.get("id")

        # validate etc.
        if job_id is None:
            return Response("MISSING ID", mimetype="text/plain", status=400)

        if job_id not in archive_store:
            return Response("UNKNOWN ID", mimetype="text/plain", status=404)

        return (
            jsonify(
                dict(
                    # this converts from DictProxy to regular dict
                    archive_store[job_id].get(
                        "progress", {"status": "unknown"}
                    )
                )
            ),
            200,
        )

    @app.route("/archive", methods=["GET"])
    @login_required(config.PASSWORD)
    def get_archive():
        """Returns result of an archive-job if ready."""
        job_id = request.args.get("id")
        # if forStatus, client only wants to validate request
        for_status = request.args.get("forStatus") is not None

        # validate etc.
        if job_id is None:
            return Response("MISSING ID", mimetype="text/plain", status=400)

        if job_id not in archive_store:
            return Response("UNKNOWN ID", mimetype="text/plain", status=404)

        if (
            archive_store[job_id]["process"].is_alive()
            or archive_store[job_id]["progress"]["status"] != "completed"
        ):
            return Response("NOT YET", mimetype="text/plain", status=202)

        archive = (
            tmp_dir / job_id / archive_store[job_id]["progress"]["artifact"]
        )

        if not archive.is_file():
            return Response(
                "FILE NOT FOUND", mimetype="text/plain", status=404
            )

        # return result
        if for_status:
            return Response("OK", mimetype="text/plain", status=200)
        return send_file(archive, as_attachment=True)

    @app.route("/archive", methods=["DELETE"])
    @login_required(config.PASSWORD)
    def delete_archive():
        """Clean up created archive or abort creation."""
        job_id = request.args.get("id")

        # validate etc.
        if job_id is None:
            return Response("MISSING ID", mimetype="text/plain", status=400)

        if job_id not in archive_store:
            return Response("UNKNOWN ID", mimetype="text/plain", status=404)

        # abort and delete artifacts
        if archive_store[job_id]["process"].is_alive():
            archive_store[job_id]["process"].kill()
            archive_store[job_id]["process"].join()

        rmtree(tmp_dir / job_id)
        return Response("DELETED", mimetype="text/plain", status=200)
